/*
 * ilSTDMS.cpp
 *
 *  Created on: 1 juin 2017
 *      Author: Mickael Ribardiere
 */

#include <mitsuba/core/fresolver.h>
#include <mitsuba/render/bsdf.h>
#include <mitsuba/hw/basicshader.h>
#include <mitsuba/core/warp.h>
#include "ior.h"
#include "microfacetSTD.h"

MTS_NAMESPACE_BEGIN


// === Interfaced Lambertian facet model (exact version): with exact STD distribution and Smith GAF
// === with or without multiple scattering using the Heitz et al. paper - Multiple-Scattering Microfacet BSDFs with the Smith Model - ACM SIGGRAPH 2016
// === For detail, see the paper:
// === "Multiple-Scattering Microfacet BSDFs with the Smith Model" - EG Workshop on Manufacturing Appearance Materials - 2017 - M. Ribardière, D. Meneveaux, B. Bringier, L. Simonot

// === The code here is directly inspired from the Heitz et al. code (https://drive.google.com/open?id=0BzvWIdpUpRx_S2UtbGNyZWtnV1U)


static inline float generateRandomNumber()
{
	const float U = ((float)rand()) / (float)RAND_MAX;
	return U;
}

static inline bool IsFiniteNumber(float x)
{
	return (x <= std::numeric_limits<Float>::max() && x >= -std::numeric_limits<Float>::max());
}

#define vec3 Vector
#define vec2 Vector

struct RayInfo
{
	// === direction
	vec3 w;
	float theta;
	float cosTheta;
	float sinTheta;
	float tanTheta;
	float alpha;
	float Lambda;

	void updateDirection(const vec3& w, const float alpha_x, const float alpha_y, const MicrofacetDistributionSTD &distrib)
	{
		this->w = w;
		theta = acosf(w.z);
		cosTheta = w.z;
		sinTheta = sinf(theta);
		tanTheta = sinTheta / cosTheta;
		const float invSinTheta2 = 1.0f / (1.0f - w.z*w.z);
		const float cosPhi2 = w.x*w.x*invSinTheta2;
		const float sinPhi2 = w.y*w.y*invSinTheta2;
		alpha = sqrtf( cosPhi2*alpha_x*alpha_x + sinPhi2*alpha_y*alpha_y );
		// === Lambda
		if(w.z > 0.9999f)
			Lambda = 0.0f;
		else if(w.z < -0.9999f)
			Lambda = -1.0f;
		else
		{
			if(w.z < 0.0)
				Lambda = -Lambda-1.0;
		}
	}

	// === height
	float h;
	float C1;
	float G1;

	void updateHeight(const float& h)
	{
		this->h = h;
		C1 = std::min(1.0f, std::max(0.0f, 0.5f*(h+1.0f)));

		if(this->w.z > 0.9999f)
			G1 = 1.0f;
		else if(this->w.z <= 0.0f)
			G1 = 0.0f;
		else
			G1 = powf(this->C1, this->Lambda);
	}
};

inline float invC1(const float U)
{
	const float h = std::max(-1.0f, std::min(1.0f, 2.0f*U-1.0f));
	return h;
}

inline float sampleHeight(const RayInfo& ray, const float U)
{
	if(ray.w.z > 0.9999f)
		return std::numeric_limits<Float>::max();
	if(ray.w.z < -0.9999f)
	{
		const float value = invC1(U*ray.C1);
		return value;
	}
	if(fabsf(ray.w.z) < 0.0001f)
		return ray.h;

	// === probability of intersection
	if (U > 1.0f - ray.G1) // === leave the microsurface
		return std::numeric_limits<Float>::max();

	const float h = invC1(
			ray.C1 / powf((1.0f-U),1.0f/ray.Lambda)
			);
	return h;
}

// === local epsilon value
#define s_EPSILON 1e-6

class ilSTDMS : public BSDF {
public:
	ilSTDMS(const Properties &props) : BSDF(props) {

		m_kd = props.getSpectrum("diffuseReflectance", Spectrum(0.5f));
		m_diffuseReflectance = new ConstantSpectrumTexture(m_kd);

		m_ni = props.getFloat("ni", 1.f);
		m_alpha = props.getFloat("alpha");

		m_gamma = props.getFloat("gamma",2);

		m_alpha2 = m_alpha*m_alpha;

		// === Here, it is possible to use the STD approximate version: change EStudentT with EStudentT_approx
		m_distrib = new MicrofacetDistributionSTD(
							MicrofacetDistributionSTD::EStudentT,
							m_alpha,
							m_alpha,
							true, m_gamma, true);

		m_invNi = 1.f/m_ni;
		m_r10 = this->r10();

		// === scattering order
		m_scatteringOrderMax = props.getInteger("scatteringOrderMax", 10);
	}

	ilSTDMS(Stream *stream, InstanceManager *manager)
	 : BSDF(stream, manager) {
		m_alpha = stream->readFloat();
		m_gamma = stream->readFloat();
		m_kd = Spectrum(stream);
		m_diffuseReflectance = static_cast<Texture *>(manager->getInstance(stream));
		m_ni = stream->readFloat();
		m_r10 = stream->readFloat();
		m_scatteringOrderMax = stream->readInt();


		m_invNi = 1.f/m_ni;
		m_alpha2 = m_alpha*m_alpha;

		// === Here again, it is possible to use the STD approximate version: change EStudentT with EStudentT_approx
		m_distrib = new MicrofacetDistributionSTD(
							MicrofacetDistributionSTD::EStudentT,
							m_alpha,
							m_alpha,
							true, m_gamma, true);

		configure();
	}

	void serialize(Stream *stream, InstanceManager *manager) const {
		BSDF::serialize(stream, manager);

		stream->writeFloat(m_alpha);
		stream->writeFloat(m_gamma);
		m_kd.serialize(stream);
		manager->serialize(stream, m_diffuseReflectance.get());
		stream->writeFloat(m_ni);
		stream->writeFloat(m_r10);
		stream->writeInt(m_scatteringOrderMax);
	}

	void configure() {
		unsigned int extraFlags = 0;

		if (!m_diffuseReflectance->isConstant())
			extraFlags |= ESpatiallyVarying;

		m_components.clear();
		m_components.push_back(EGlossyReflection | EFrontSide | extraFlags);

		/* Verify the input parameters and fix them if necessary */
		m_diffuseReflectance = ensureEnergyConservation(
				m_diffuseReflectance, "diffuseReflectance", 1.0f);

		m_usesRayDifferentials = m_diffuseReflectance->usesRayDifferentials();

		BSDF::configure();
	}

	Spectrum eval(const BSDFSamplingRecord &bRec, EMeasure measure) const {
		/* Stop if this component was not requested */
		if (measure != ESolidAngle ||
			Frame::cosTheta(bRec.wi) <= 0 ||
			Frame::cosTheta(bRec.wo) <= 0 ||
			((bRec.component != -1 && bRec.component != 0) ||
			!(bRec.typeMask & EGlossyReflection)))
			return Spectrum(0.0f);
		
		vec3 wi(bRec.wi.x, bRec.wi.y, bRec.wi.z);
		vec3 wo(bRec.wo.x, bRec.wo.y, bRec.wo.z);

		// === start random walk from either wi or wo randomly (bidirectional path tracing)
		Spectrum res = (generateRandomNumber() > 0.5f) ?
			2.0f * eval_IL(wi, wo, m_alpha, m_alpha, m_scatteringOrderMax) :
			2.0f * eval_IL(wo, wi, m_alpha, m_alpha, m_scatteringOrderMax)/Frame::cosTheta(bRec.wi)*Frame::cosTheta(bRec.wo);

		return res;
	}

	Float pdf(const BSDFSamplingRecord &bRec, EMeasure measure) const {
		bool hasSpecular = (bRec.typeMask & EGlossyReflection) &&
			(bRec.component == -1 || bRec.component == 0);
		bool hasDiffuse = (bRec.typeMask & EDiffuseReflection) &&
			(bRec.component == -1 || bRec.component == 1);

		if (measure != ESolidAngle ||
			Frame::cosTheta(bRec.wi) <= 0 ||
			Frame::cosTheta(bRec.wo) <= 0 ||
			(!hasSpecular && !hasDiffuse))
			return 0.0f;

		Float diff = m_diffuseReflectance->eval(bRec.its).average();

		Vector H = normalize(bRec.wo+bRec.wi);

		// === Jacobian of the half-direction mapping
		const Float dwh_dwo = 1.0f / (4.0f * absDot(bRec.wo, H));
		// === Evaluate the microfacet model sampling density function
		const Float prob = m_distrib->pdf(bRec.wi, H);
		Float specPdf = prob * dwh_dwo;

		// single-scattering PDF + diffuse 
		// otherwise too many fireflies due to lack of multiple-scattering PDF
		// (MIS works even if the PDF is wrong and not normalized)
		return diff*(warp::squareToCosineHemispherePdf(bRec.wo)) + (1.f-diff)*specPdf;
	}

	Spectrum sample(BSDFSamplingRecord &bRec, const Point2 &sample) const {
		if (Frame::cosTheta(bRec.wi) < 0 ||
			((bRec.component != -1 && bRec.component != 0) ||
			!(bRec.typeMask & EGlossyReflection)))
			return Spectrum(0.0f);

		Float pdf;
		return this->sample(bRec, pdf, sample);
	}

	Spectrum sample(BSDFSamplingRecord &bRec, Float &pdf, const Point2 &sample) const {
		if (Frame::cosTheta(bRec.wi) < 0 ||
			((bRec.component != -1 && bRec.component != 0) ||
			!(bRec.typeMask & EGlossyReflection)))
			return Spectrum(0.0f);

		bRec.eta = 1.0f;
		bRec.sampledComponent = 0;
		bRec.sampledType = EGlossyReflection;	

		vec3 wi(bRec.wi.x, bRec.wi.y, bRec.wi.z);

		Spectrum energy;
		vec3 wo = sample_IL(wi, m_alpha, m_alpha, m_scatteringOrderMax, energy);
		bRec.wo.x = wo.x;
		bRec.wo.y = wo.y;
		bRec.wo.z = wo.z;

		pdf = this->pdf(bRec, ESolidAngle);
		
		return energy;
	}

	void addChild(const std::string &name, ConfigurableObject *child) {
		if (child->getClass()->derivesFrom(MTS_CLASS(Texture))) {
			if (name == "diffuseReflectance")
				m_diffuseReflectance = static_cast<Texture *>(child);
			else
				BSDF::addChild(name, child);
		} else {
			BSDF::addChild(name, child);
		}
	}

	Float getRoughness(const Intersection &its, int component) const {
		return m_alpha;
	}

	std::string toString() const {
		std::ostringstream oss;
		oss << "ilSTDMS[" << endl
			<< "  id = \"" << getID() << "\"," << endl
			<< "  id = \"" << getID() << "\"," << endl
			<< "  alpha = " << m_alpha << "," << endl
			<< "  diffuseReflectance = " << indent(m_diffuseReflectance->toString()) << "," << endl
			<< "  ni = " << m_ni << "," << endl
			<< "]";
		return oss.str();
	}

	Shader *createShader(Renderer *renderer) const;

	MTS_DECLARE_CLASS()
private:
	ref<Texture> m_diffuseReflectance;
	Float m_alpha;
	Float m_ni;
	Float m_r10;

	ref<MicrofacetDistributionSTD> m_distrib;
	Float m_gamma;

	Spectrum m_kd;

	Float m_invNi;
	Float m_alpha2;

	int m_scatteringOrderMax;

protected:

	Float r10() const {
		if (m_ni<=(1.0f+s_EPSILON)) return 0.0f;
		return 1.0-((m_invNi*m_invNi)*(1.0-this->r01()));
	}

	Float r01() const {
		if (m_ni<=(1.0f+s_EPSILON)) return 0.0f;

		Float n=m_ni, n2=n*n;
		Float n3=n2*n, n4=n2*n2;
		Float n2p1=(n2+1),n2m1=(n2-1);
		Float n2p12=n2p1*n2p1, n2m12=n2m1*n2m1;

		// MOLENAAR formulation
		Float T1 = 2.0*n3*(n2+2*n-1);
		Float B1=n2p1*(n4-1);

		Float T2=(n-1)*(3*n+1);
		Float B2=6*(n+1)*(n+1);

		Float T3=8.0*n4*(n4+1);
		Float B3=(n2+1)*(n4-1)*(n4-1);

		Float T4=n2*n2m12;
		Float B4=n2p12*n2p1;

		return (1.0/2.0-T1/B1+T2/B2+T3/B3*math::fastlog(n)-T4/B4*math::fastlog((n+1)/(n-1)));
	}

	Float T01(const Vector &dir, const Vector &H) const {
		if (m_ni<=(1.0f+s_EPSILON)) return 1.0f;

		Float c1 = fabs(dot(dir,H));
		Float r = 0.f;
		Float c2 = 1.f - m_invNi*m_invNi * (1.f - c1*c1);

		if(c2 > 0.0) {     // === refracted ray
			c2 = sqrt(c2);
			Float Rs = (m_invNi*c1 - c2) / (m_invNi*c1 + c2);   // === s-polarised fresnel reflectance
			Float Rp = (m_invNi*c2 - c1) / (m_invNi*c2 + c1);   // === p-polarised fresnel reflectance
			Rs *= Rs;
			Rp *= Rp;
			r = (Rs + Rp) * 0.5f;            // === fresnel reflectance
		}
		return 1.f-r;
	}

	//=================================================================
	// Fresnel coefficient
	//=================================================================
	Float FresnelNi(const Vector &N, const Vector &V) const	{
		Float c=dot(N,V);
		if(c<0.0) return 0.0;
		Float g=sqrt(m_ni*m_ni+c*c-1.0f);
		if(g==0.f)
			return 0.f;
		Float gmc=g-c, gpc=g+c;
		Float cgpc=c*gpc-1.0, cgmc=c*gmc+1.0;
		Float F=0.5*(gmc*gmc)/(gpc*gpc)*(1.0+(cgpc*cgpc)/(cgmc*cgmc));
		return F;
	}

	Spectrum rSimoSTDTSfr(const Vector &rVi, const Vector &rVo, const Float &G2, const Spectrum &kd) const {
		Float cto = Frame::cosTheta(rVo);
		Float cti = Frame::cosTheta(rVi);

		// === SPECULAR PART
		Spectrum SPEC(0.f);
		if(cto>=s_EPSILON && cti>=s_EPSILON) {
			Vector Hspec=(rVo+rVi);
			Hspec = normalize(Hspec);

			Float dotProd = dot(rVi,Hspec);

			if(dotProd>=0.0) {
				// === Compute the GAF
				Float Gspec = G2; //m_distrib->G(rVi, rVo, Hspec);

				// === Compute Fresnel
				Float g=sqrt(m_ni*m_ni+dotProd*dotProd-1.0f);
				Float gmc=g-dotProd, gpc=g+dotProd;
				Float cgpc=dotProd*gpc-1.0, cgmc=dotProd*gmc+1.0;
				Float F = 0.5*(gmc*gmc)/(gpc*gpc)*(1.0+(cgpc*cgpc)/(cgmc*cgmc));

				SPEC = Spectrum((F*m_distrib->eval(Hspec)*Gspec)/(4.0*(cti)*(cto)));
			}
		}

		// === DIFFUSE PART
		// === choose a ramdom H vector
		Point2	sample(generateRandomNumber(), generateRandomNumber());
		//Point2	sample(0.5,0.5);
		Float pdfDiff;
		Vector Hdiff = m_distrib->sample(rVi, sample, pdfDiff);
		if(pdfDiff <= 0.0)
			return SPEC*cto;


		// === compute the bsdf for the chosen micro-facet with Hdiff normal
		Float frm=INV_PI*m_invNi*m_invNi;
		frm*=this->T01(rVi,Hdiff)*this->T01(rVo,Hdiff);
		Spectrum den = Spectrum(1.f) - kd*m_r10;
		Spectrum DIFF((kd/den)*frm);

		// === compute the GAF
		Float Gdiff = 0.0;
		Float dotVi = dot(rVi,Hdiff);
		Float dotVo = dot(rVo,Hdiff);
		Gdiff = G2;

		dotVi = (dotVi<0)?0.0:dotVi;
		dotVo = (dotVo<0)?0.0:dotVo;
		DIFF *= Gdiff*dotVo*dotVi;

		Float SinSon = cti*cto;
		SinSon = (SinSon<(s_EPSILON))?(s_EPSILON):SinSon;
		DIFF /= SinSon;

		DIFF *= m_distrib->eval(Hdiff) / pdfDiff;

		return (SPEC+DIFF) * cto;
	}


	// === MULTIPLE SCATTERING ROUTINES

	vec3 samplePhaseFunction_IL(const vec3& wi, Spectrum& weight) const
	{
		float U1 = generateRandomNumber();
		float U2 = generateRandomNumber();

		vec3 wm = m_distrib->sample(wi, Point2(U1, U2));

		vec3 wo;

		Frame localFrame(wm);

		// === use Fresnel coefficient for specular sampling
		Spectrum coeffFresnel(this->FresnelNi(localFrame.n, wi));
		Float specularWeight = coeffFresnel.average();

		weight = Spectrum(0.0);

		// === first, choose specular or smooth part of the brdf
		Float R1 = (double) rand() / (RAND_MAX);
		if (R1 < specularWeight) {
			// === specular component
			wo = reflect(wi, localFrame.n);
			wo = normalize(wo);

			Float dotProd = dot(localFrame.n, wo);
			if (dotProd > 0)
				weight = coeffFresnel / (specularWeight);
		} else {
			Point2 sample(generateRandomNumber(),generateRandomNumber());
			wo = warp::squareToCosineHemisphere(sample);
			wo = localFrame.toWorld(wo);

			Float frm=INV_PI*m_invNi*m_invNi;
			frm*=this->T01(wi,localFrame.n)*this->T01(wo,localFrame.n);
			Spectrum den = Spectrum(1.f) - m_kd*m_r10;
			weight = ((m_kd/den)*frm);

			if (!weight.isZero()) {
				Float pdf = (1.0f - specularWeight) * INV_PI;
				if(pdf!=0)
					weight /= pdf;
			}
		}

		return wo;
	}

	Spectrum evalPhaseFunction_IL(const RayInfo& ray, const vec3& wo, const float alpha_x, const float alpha_y) const
	{
		if(ray.w.z > 0.9999f)
			return Spectrum(0.0f);

		// half vector
		const vec3 wh = normalize(-ray.w+wo);
		if(wh.z < 0.0f)
			return Spectrum(0.0f);

		// projected area
		float projectedArea;
		if(ray.w.z < -0.9999f)
			projectedArea = 1.0f;
		else
			projectedArea = ray.Lambda * ray.w.z;

		// value
		// === Compute Fresnel
		Float dotProd = dot(-ray.w,wh);
		Float g=sqrt(m_ni*m_ni+dotProd*dotProd-1.0f);
		Float gmc=g-dotProd, gpc=g+dotProd;
		Float cgpc=dotProd*gpc-1.0, cgmc=dotProd*gmc+1.0;
		Float F = 0.5*(gmc*gmc)/(gpc*gpc)*(1.0+(cgpc*cgpc)/(cgmc*cgmc));

		Spectrum SPEC = Spectrum(m_distrib->eval(wh) * F * std::max(0.0f, dot(-ray.w, wh)) / 4.0f / projectedArea / dot(-ray.w, wh));


		// === DIFFUSE PART
		// === choose a ramdom H vector
		Point2	sample(generateRandomNumber(), generateRandomNumber());
		//Point2	sample(0.5,0.5);
		Float pdfDiff;
		Vector Hdiff = m_distrib->sample(-ray.w, sample, pdfDiff);
		if(pdfDiff <= 0.0)
			return SPEC;//*cto;


		// === compute the bsdf for the chosen micro-facet with Hdiff normal
		Float frm=INV_PI*m_invNi*m_invNi;
		frm*=this->T01(-ray.w,Hdiff)*this->T01(wo,Hdiff);
		Spectrum den = Spectrum(1.f) - m_kd*m_r10;
		Spectrum DIFF((m_kd/den)*frm);

		DIFF *= m_distrib->eval(Hdiff) * std::max(0.0f, dot(wo, Hdiff));
		DIFF /= pdfDiff;

		return (SPEC+DIFF);
	}

	// === MIS weights for bidirectional path tracing on the microsurface
	float MISweight_IL(const vec3& wi, const vec3& wo, const float alpha_x, const float alpha_y) const
	{
		if(wi.x == -wo.x && wi.y == -wo.y && wi.z == -wo.z)
			return 1.0f;
		const vec3 wh = normalize(wi+wo);
		const float value = m_distrib->eval( (wh.z>0) ? wh : -wh );
		return value;
	}

	Spectrum eval_IL(const vec3& wi, const vec3& wo, const float alpha_x, const float alpha_y, const int scatteringOrderMax) const
	{
		if(wi.z <= 0 || wo.z <= 0)
			return Spectrum(0.0f);

		// === init
		RayInfo ray;
		ray.updateDirection(-wi, alpha_x, alpha_y, *m_distrib.get());
		ray.updateHeight(1.0f);
		Spectrum energy(1.0f);

		RayInfo ray_shadowing;
		ray_shadowing.updateDirection(wo, alpha_x, alpha_y, *m_distrib.get());

		// === eval single scattering
		const float G2 = 1.0f / (1.0f + (-ray.Lambda-1.0f) + ray_shadowing.Lambda);
		Spectrum singleScattering = rSimoSTDTSfr(wi, wo, G2, m_kd);


		// === MIS weight
		float wi_MISweight = 0.0;;

		// === multiple scattering
		Spectrum multipleScattering(0.0f);

		// === random walk
		int current_scatteringOrder = 0;
		while(current_scatteringOrder < scatteringOrderMax)
		{
			// === next height
			float U = generateRandomNumber();
			ray.updateHeight( sampleHeight(ray, U) );

			// === leave the microsurface?
			if( ray.h == std::numeric_limits<Float>::max() )
				break;
			else
				current_scatteringOrder++;

			// === next event estimation
			if( current_scatteringOrder > 1) // === single scattering is already computed
			{
				Spectrum phasefunction = evalPhaseFunction_IL(ray, wo, alpha_x, alpha_y);
				ray_shadowing.updateHeight(ray.h);
				float shadowing = ray_shadowing.G1;
				Spectrum I = energy * phasefunction * shadowing;

				// === MIS
				const float MIS = wi_MISweight / ( wi_MISweight + MISweight_IL(-ray.w, wo, alpha_x, alpha_y) );


				if ( IsFiniteNumber(I[0]) )
					multipleScattering += I * MIS;
			}

			// === next direction
			Spectrum weight;
			ray.updateDirection(samplePhaseFunction_IL(-ray.w, weight), alpha_x, alpha_y, *m_distrib.get());
			energy = energy * weight;
			ray.updateHeight(ray.h);

			if(current_scatteringOrder == 1)
				wi_MISweight = MISweight_IL(wi, ray.w, alpha_x, alpha_y);

			// === if NaN (should not happen, just in case)
			if( (ray.h != ray.h) || (ray.w.x != ray.w.x))
				return Spectrum(0.0f);
		}

		// === 0.5f = MIS weight of singleScattering
		// === multipleScattering already weighted by MIS
		return 0.5f*singleScattering + multipleScattering;
	}

	vec3 sample_IL(const vec3& wi, const float alpha_x, const float alpha_y, const int scatteringOrderMax, Spectrum& energy) const
	{
		energy = Spectrum(1.0f);

		// === init
		RayInfo ray;
		ray.updateDirection(-wi, alpha_x, alpha_y, *m_distrib.get());
		ray.updateHeight(1.0f);

		// === random walk
		int current_scatteringOrder = 0;
		while(true)
		{
			// === next height
			float U = generateRandomNumber();
			ray.updateHeight( sampleHeight(ray, U) );

			// leave the microsurface?
			if( ray.h == std::numeric_limits<Float>::max() )
				break;
			else
				current_scatteringOrder++;

			// === next direction
			Spectrum weight;
			ray.updateDirection(samplePhaseFunction_IL(-ray.w, weight), alpha_x, alpha_y, *m_distrib.get());
			energy = energy * weight;
			ray.updateHeight(ray.h);

			// === if NaN (should not happen, just in case)
			if( (ray.h != ray.h) || (ray.w.x != ray.w.x))
			{
				energy = Spectrum(0.0f);
				return vec3(0,0,1);
			}

			if( current_scatteringOrder > scatteringOrderMax )
			{
				energy = Spectrum(0.0f);
				return vec3(0,0,1);
			}
		}

		return ray.w;
	}


};

/**
 * GLSL port of the rough conductor shader. This version is much more
 * approximate -- it only supports the Ashikhmin-Shirley distribution,
 * does everything in RGB, and it uses the Schlick approximation to the
 * Fresnel reflectance of conductors. When the roughness is lower than
 * \alpha < 0.2, the shader clamps it to 0.2 so that it will still perform
 * reasonably well in a VPL-based preview.
 */
class ilSTDMSShader : public Shader {
public:
	ilSTDMSShader(Renderer *renderer, const Texture *specularReflectance,
			const Texture *alphaU, const Texture *alphaV, const Spectrum &eta,
			const Spectrum &k) : Shader(renderer, EBSDFShader),
			m_specularReflectance(specularReflectance), m_alphaU(alphaU), m_alphaV(alphaV){
		m_specularReflectanceShader = renderer->registerShaderForResource(m_specularReflectance.get());
		m_alphaUShader = renderer->registerShaderForResource(m_alphaU.get());
		m_alphaVShader = renderer->registerShaderForResource(m_alphaV.get());

		/* Compute the reflectance at perpendicular incidence */
		m_R0 = fresnelConductorExact(1.0f, eta, k);
	}

	bool isComplete() const {
		return m_specularReflectanceShader.get() != NULL &&
			   m_alphaUShader.get() != NULL &&
			   m_alphaVShader.get() != NULL;
	}

	void putDependencies(std::vector<Shader *> &deps) {
		deps.push_back(m_specularReflectanceShader.get());
		deps.push_back(m_alphaUShader.get());
		deps.push_back(m_alphaVShader.get());
	}

	void cleanup(Renderer *renderer) {
		renderer->unregisterShaderForResource(m_specularReflectance.get());
		renderer->unregisterShaderForResource(m_alphaU.get());
		renderer->unregisterShaderForResource(m_alphaV.get());
	}

	void resolve(const GPUProgram *program, const std::string &evalName, std::vector<int> &parameterIDs) const {
		parameterIDs.push_back(program->getParameterID(evalName + "_R0", false));
	}

	void bind(GPUProgram *program, const std::vector<int> &parameterIDs, int &textureUnitOffset) const {
		program->setParameter(parameterIDs[0], m_R0);
	}

	void generateCode(std::ostringstream &oss,
			const std::string &evalName,
			const std::vector<std::string> &depNames) const {
		oss << "uniform vec3 " << evalName << "_R0;" << endl
			<< endl
			<< "float " << evalName << "_D(vec3 m, float alphaU, float alphaV) {" << endl
			<< "    float ct = cosTheta(m), ds = 1-ct*ct;" << endl
			<< "    if (ds <= 0.0)" << endl
			<< "        return 0.0f;" << endl
			<< "    alphaU = 2 / (alphaU * alphaU) - 2;" << endl
			<< "    alphaV = 2 / (alphaV * alphaV) - 2;" << endl
			<< "    float exponent = (alphaU*m.x*m.x + alphaV*m.y*m.y)/ds;" << endl
			<< "    return sqrt((alphaU+2) * (alphaV+2)) * 0.15915 * pow(ct, exponent);" << endl
			<< "}" << endl
			<< endl
			<< "float " << evalName << "_G(vec3 m, vec3 wi, vec3 wo) {" << endl
			<< "    if ((dot(wi, m) * cosTheta(wi)) <= 0 || " << endl
			<< "        (dot(wo, m) * cosTheta(wo)) <= 0)" << endl
			<< "        return 0.0;" << endl
			<< "    float nDotM = cosTheta(m);" << endl
			<< "    return min(1.0, min(" << endl
			<< "        abs(2 * nDotM * cosTheta(wo) / dot(wo, m))," << endl
			<< "        abs(2 * nDotM * cosTheta(wi) / dot(wi, m))));" << endl
			<< "}" << endl
			<< endl
			<< "vec3 " << evalName << "_schlick(float ct) {" << endl
			<< "    float ctSqr = ct*ct, ct5 = ctSqr*ctSqr*ct;" << endl
			<< "    return " << evalName << "_R0 + (vec3(1.0) - " << evalName << "_R0) * ct5;" << endl
			<< "}" << endl
			<< endl
			<< "vec3 " << evalName << "(vec2 uv, vec3 wi, vec3 wo) {" << endl
			<< "   if (cosTheta(wi) <= 0 || cosTheta(wo) <= 0)" << endl
			<< "    	return vec3(0.0);" << endl
			<< "   vec3 H = normalize(wi + wo);" << endl
			<< "   vec3 reflectance = " << depNames[0] << "(uv);" << endl
			<< "   float alphaU = max(0.2, " << depNames[1] << "(uv).r);" << endl
			<< "   float alphaV = max(0.2, " << depNames[2] << "(uv).r);" << endl
			<< "   float D = " << evalName << "_D(H, alphaU, alphaV)" << ";" << endl
			<< "   float G = " << evalName << "_G(H, wi, wo);" << endl
			<< "   vec3 F = " << evalName << "_schlick(1-dot(wi, H));" << endl
			<< "   return reflectance * F * (D * G / (4*cosTheta(wi)));" << endl
			<< "}" << endl
			<< endl
			<< "vec3 " << evalName << "_diffuse(vec2 uv, vec3 wi, vec3 wo) {" << endl
			<< "    if (cosTheta(wi) < 0.0 || cosTheta(wo) < 0.0)" << endl
			<< "    	return vec3(0.0);" << endl
			<< "    return " << evalName << "_R0 * inv_pi * inv_pi * cosTheta(wo);"<< endl
			<< "}" << endl;
	}
	MTS_DECLARE_CLASS()
private:
	ref<const Texture> m_specularReflectance;
	ref<const Texture> m_alphaU;
	ref<const Texture> m_alphaV;
	ref<Shader> m_specularReflectanceShader;
	ref<Shader> m_alphaUShader;
	ref<Shader> m_alphaVShader;
	Spectrum m_R0;
};

Shader *ilSTDMS::createShader(Renderer *renderer) const {
	return new ilSTDMSShader(renderer,
		new ConstantSpectrumTexture(Spectrum(1.0)), new ConstantFloatTexture(m_alpha), new ConstantFloatTexture(m_alpha), Spectrum(1.0), Spectrum(1.0));
}

MTS_IMPLEMENT_CLASS(ilSTDMSShader, false, Shader)
MTS_IMPLEMENT_CLASS_S(ilSTDMS, false, BSDF)
MTS_EXPORT_PLUGIN(ilSTDMS, "IL BRDF model with STD and multiple scattering");
MTS_NAMESPACE_END
