/*
 * ilstdexact.cpp
 *
 *  Created on: 15 mai 2017
 *      Author: Mickael Ribardiere
 */



#include <mitsuba/render/bsdf.h>
#include <mitsuba/hw/basicshader.h>
#include <mitsuba/core/warp.h>
#include "microfacetSTD.h"
#include "rtrans.h"
#include "ior.h"

MTS_NAMESPACE_BEGIN

// === Interfaced Lambertian facet model (exact version): with approx STD distribution and Torrance-Sparrow or Smith GAF
// === with or without multiple scattering in V-cavities (Torrance-Sparrow GAF)
// === For detail, see the paper:
// === "Rendering Rough Opaque Materials with Interfaced Lambertian Microfacets" - IEEE TVCG - Meneveaux D., Bringier B., Tauzia E., Ribardiere M., Simonot L. - 2017

// === local epsilon value
#define s_EPSILON 1e-6

class ILSTD_exact : public BSDF {
public:
	ILSTD_exact(const Properties &props) : BSDF(props) {

		m_diffuseReflectance = new ConstantSpectrumTexture(
			props.getSpectrum("diffuseReflectance", Spectrum(0.5f)));

		m_ni = props.getFloat("ni", 1.f);
		m_alpha = props.getFloat("alpha");

		m_gamma = props.getFloat("gamma");

		m_maxDepth = props.getInteger("maxDepth", -1); // === 1 means no inter-reflexions, 2 means 2 bounces... -1 means infinity bounces
		m_nbInterSamples = props.getInteger("nbInterSamples", 1);

		m_smithGAF = props.getBoolean("smith",false);


		m_alpha2 = m_alpha*m_alpha;

		m_sampleVisible = props.getBoolean("sampleVisible",true);

		m_correlated = props.getBoolean("correlated",false);

		m_distrib = new MicrofacetDistributionSTD(
							MicrofacetDistributionSTD::EStudentT_approx,
							m_alpha,
							m_alpha,
							m_sampleVisible, m_gamma, m_correlated);

		m_invNi = 1.f/m_ni;
		m_r10 = this->r10();
	}

	ILSTD_exact(Stream *stream, InstanceManager *manager)
	 : BSDF(stream, manager) {
		m_alpha = stream->readFloat();
		m_gamma = stream->readFloat();
		m_diffuseReflectance = static_cast<Texture *>(manager->getInstance(stream));
		m_ni = stream->readFloat();
		m_r10 = stream->readFloat();
		m_maxDepth = stream->readInt();
		m_nbInterSamples = stream->readInt();
		m_smithGAF = stream->readBool();
		m_sampleVisible = stream->readBool();
		m_correlated = stream->readBool();

		m_invNi = 1.f/m_ni;
		m_alpha2 = m_alpha*m_alpha;

		m_distrib = new MicrofacetDistributionSTD(
							MicrofacetDistributionSTD::EStudentT_approx,
							m_alpha,
							m_alpha,
							m_sampleVisible, m_gamma, m_correlated);

		configure();
	}

	void serialize(Stream *stream, InstanceManager *manager) const {
		BSDF::serialize(stream, manager);

		stream->writeFloat(m_alpha);
		stream->writeFloat(m_gamma);
		manager->serialize(stream, m_diffuseReflectance.get());
		stream->writeFloat(m_ni);
		stream->writeFloat(m_r10);
		stream->writeInt(m_maxDepth);
		stream->writeInt(m_nbInterSamples);
		stream->writeBool(m_smithGAF);
		stream->writeBool(m_sampleVisible);
		stream->writeBool(m_correlated);
	}

	void configure() {
		unsigned int extraFlags = 0;

		if (!m_diffuseReflectance->isConstant())
			extraFlags |= ESpatiallyVarying;

		m_components.clear();
		m_components.push_back(EGlossyReflection | EFrontSide | extraFlags);

		/* Verify the input parameters and fix them if necessary */
		m_diffuseReflectance = ensureEnergyConservation(
				m_diffuseReflectance, "diffuseReflectance", 1.0f);

		m_usesRayDifferentials = m_diffuseReflectance->usesRayDifferentials();

		BSDF::configure();
	}

	Spectrum getDiffuseReflectance(const Intersection &its) const {
		return m_diffuseReflectance->eval(its);
	}

	Spectrum getSpecularReflectance(const Intersection &its) const {
		return Spectrum(1.f);
	}

	/// Helper function: reflect \c wi with respect to a given surface normal
	inline Vector reflect(const Vector &wi, const Normal &m) const {
		return 2 * dot(wi, m) * Vector(m) - wi;
	}

	Spectrum eval(const BSDFSamplingRecord &bRec, EMeasure measure) const {
		if (Frame::cosTheta(bRec.wi) <= 0 ||
			Frame::cosTheta(bRec.wo) <= 0 || measure != ESolidAngle)
			return Spectrum(0.0f);

		return this->rSimoSTDTSfr(bRec.wi, bRec.wo, m_diffuseReflectance->eval(bRec.its), bRec.its);
	}

	Float pdf(const BSDFSamplingRecord &bRec, EMeasure measure) const {
		bool hasSpecular = (bRec.typeMask & EGlossyReflection) &&
			(bRec.component == -1 || bRec.component == 0);
		bool hasDiffuse = (bRec.typeMask & EDiffuseReflection) &&
			(bRec.component == -1 || bRec.component == 1);

		if (measure != ESolidAngle ||
			Frame::cosTheta(bRec.wi) <= 0 ||
			Frame::cosTheta(bRec.wo) <= 0 ||
			(!hasSpecular && !hasDiffuse))
			return 0.0f;

		Float diff = m_diffuseReflectance->eval(bRec.its).average();

		Vector H = normalize(bRec.wo+bRec.wi);

		// === Jacobian of the half-direction mapping
		const Float dwh_dwo = 1.0f / (4.0f * absDot(bRec.wo, H));
		// === Evaluate the microfacet model sampling density function
		const Float prob = m_distrib->pdf(bRec.wi, H);
		Float specPdf = prob * dwh_dwo;

		return diff*(warp::squareToCosineHemispherePdf(bRec.wo)) + (1.f-diff)*specPdf;
	}

	inline Spectrum sample(BSDFSamplingRecord &bRec, Float &_pdf, const Point2 &_sample) const {
		if (Frame::cosTheta(bRec.wi) <= 0)
			return Spectrum(0.0f);

		Float probDiffuse = m_diffuseReflectance->eval(bRec.its).average();
		Point2 sample(_sample);

		Vector H;
		if(sample.y <= probDiffuse) { // === diffuse sampling
			sample.y /= probDiffuse;
			bRec.sampledComponent = 1;
			bRec.sampledType = EDiffuseReflection;
			bRec.wo = warp::squareToCosineHemisphere(sample);
			H=normalize(bRec.wo+bRec.wi);
		} else { // === specular sampling
			sample.y = (sample.y - probDiffuse) / (1 - probDiffuse);
			H = m_distrib->sample(bRec.wi, sample);
			bRec.wo = reflect(bRec.wi, H);
			bRec.sampledComponent = 0;
			bRec.sampledType = EGlossyReflection;
			if (Frame::cosTheta(bRec.wo) <= 0)
				return Spectrum(0.0f);
		}

		bRec.eta = 1.0f;

		_pdf = pdf(bRec, ESolidAngle);

		if(_pdf == 0)
			return Spectrum(0.f);

		return eval(bRec, ESolidAngle) / _pdf;
	}

	Spectrum sample(BSDFSamplingRecord &bRec, const Point2 &sample) const {
		Float pdf;
		return ILSTD_exact::sample(bRec, pdf, sample);
	}

	void addChild(const std::string &name, ConfigurableObject *child) {
		if (child->getClass()->derivesFrom(MTS_CLASS(Texture))) {
			if (name == "diffuseReflectance")
				m_diffuseReflectance = static_cast<Texture *>(child);
			else
				BSDF::addChild(name, child);
		} else {
			BSDF::addChild(name, child);
		}
	}

	Float getRoughness(const Intersection &its, int component) const {
		return m_alpha;
	}

	std::string toString() const {
		std::ostringstream oss;
		oss << "ILSTDTS_exact[" << endl
			<< "  id = \"" << getID() << "\"," << endl
			<< "  alpha = " << m_alpha << "," << endl
			<< "  diffuseReflectance = " << indent(m_diffuseReflectance->toString()) << "," << endl
			<< "  ni = " << m_ni << "," << endl
			<< "]";
		return oss.str();
	}

	Shader *createShader(Renderer *renderer) const;

	MTS_DECLARE_CLASS()
private:
	ref<Texture> m_diffuseReflectance;
	Float m_alpha;
	Float m_ni;
	Float m_r10;

	ref<MicrofacetDistributionSTD> m_distrib;
	Float m_gamma;

	bool m_smithGAF;

	bool m_sampleVisible;

	bool m_correlated;

	Float m_invNi;
	Float m_alpha2;

	int	m_maxDepth; // === max inter-reflection depth
	int m_nbInterSamples; // === number of samples to compute the inter-reflections component

	Spectrum rSimoSTDTSfr(const Vector &rVi, const Vector &rVo, const Spectrum &kd, const Intersection &its) const {
		Float cto = Frame::cosTheta(rVo);
		Float cti = Frame::cosTheta(rVi);

		// === SPECULAR PART
		Spectrum SPEC(0.f);
		if(cto>=s_EPSILON && cti>=s_EPSILON) {
			Vector Hspec=(rVo+rVi);
			Hspec = normalize(Hspec);

			Float dotProd = dot(rVi,Hspec);

			if(dotProd>=0.0) {
				// === Compute the GAF
				Float Gspec = 0.0;
				if(m_smithGAF) {
					Gspec = m_distrib->G(rVi, rVo, Hspec);
				} else {
					Float dotNH = Frame::cosTheta(Hspec);
					Gspec=std::min((2.0*(dotNH)*(cto)/(dotProd)),(2.0*(dotNH)*(cti)/(dotProd)));
					Gspec = std::max(0.0f, std::min(Gspec,1.0f));
				}

				// === Compute Fresnel
				Float g=sqrt(m_ni*m_ni+dotProd*dotProd-1.0f);
				Float gmc=g-dotProd, gpc=g+dotProd;
				Float cgpc=dotProd*gpc-1.0, cgmc=dotProd*gmc+1.0;
				Float F = 0.5*(gmc*gmc)/(gpc*gpc)*(1.0+(cgpc*cgpc)/(cgmc*cgmc));

				SPEC = Spectrum((F*m_distrib->eval(Hspec)*Gspec)/(4.0*(cti)*(cto)));
			}
		}

		// === DIFFUSE PART
		// === choose a ramdom H vector
		Point2	sample(((double) rand() / (RAND_MAX)),((double) rand() / (RAND_MAX)));
		//Point2	sample(0.5,0.5);
		Float pdfDiff;
		Vector Hdiff = m_distrib->sample(rVi, sample, pdfDiff);
		if(pdfDiff <= 0.0)
			return SPEC*cto;


		// === compute the bsdf for the chosen micro-facet with Hdiff normal
		Float frm=INV_PI*m_invNi*m_invNi;
		frm*=this->T01(rVi,Hdiff)*this->T01(rVo,Hdiff);
		Spectrum den = Spectrum(1.f) - kd*m_r10;
		Spectrum DIFF((kd/den)*frm);

		// === compute the GAF
		Float Gdiff = 0.0;
		Float dotVi = dot(rVi,Hdiff);
		Float dotVo = dot(rVo,Hdiff);
		Float dotNHdiff = Frame::cosTheta(Hdiff);
		if(m_smithGAF) {
			Gdiff = m_distrib->G(rVi, rVo, Hdiff);
		} else {
			Gdiff=std::min((2.0*(dotNHdiff)*(cto)/(dotVo)),(2.0*(dotNHdiff)*(cti)/(dotVi)));
			Gdiff = std::max(0.0f, std::min(Gdiff,1.0f));
		}

		dotVi = (dotVi<0)?0.0:dotVi;
		dotVo = (dotVo<0)?0.0:dotVo;
		DIFF *= Gdiff*dotVo*dotVi;///dotNHdiff;

		Float SinSon = cti*cto;
		SinSon = (SinSon<(s_EPSILON))?(s_EPSILON):SinSon;
		DIFF /= SinSon;

		if(m_sampleVisible)
			DIFF *= m_distrib->eval(Hdiff) / pdfDiff;
		else
			DIFF /= dotNHdiff;

		// === Compute the micro-facets inter-reflections
		if(m_maxDepth != 1 && !m_smithGAF) {
			Float phiH = std::atan2(Hdiff.y, Hdiff.x);
			DIFF += this->computeInterReflections(Hdiff, phiH, rVi, rVo, kd);
		}

		return (SPEC+DIFF) * cto;
	}

	Float T01(const Vector &dir, const Vector &H) const {
		if (m_ni<=(1.0f+s_EPSILON)) return 1.0f;

		Float c1 = dot(dir,H);
		Float r = 0.f;
		Float c2 = 1.f - m_invNi*m_invNi * (1.f - c1*c1);

		if(c2 > 0.0) {     // refracted ray
			c2 = sqrt(c2);
			Float Rs = (m_invNi*c1 - c2) / (m_invNi*c1 + c2);   // s-polarised fresnel reflectance
			Float Rp = (m_invNi*c2 - c1) / (m_invNi*c2 + c1);   // p-polarised fresnel reflectance
			Rs *= Rs;
			Rp *= Rp;
			r = (Rs + Rp) * 0.5f;            // fresnel reflectance
		}
		return 1.f-r;
	}

	Float r10() const {
		if (m_ni<=(1.0f+s_EPSILON)) return 0.0f;
		return 1.0-((m_invNi*m_invNi)*(1.0-this->r01()));
	}

	Float r01() const {
		if (m_ni<=(1.0f+s_EPSILON)) return 0.0f;

		Float n=m_ni, n2=n*n;
		Float n3=n2*n, n4=n2*n2;
		Float n2p1=(n2+1),n2m1=(n2-1);
		Float n2p12=n2p1*n2p1, n2m12=n2m1*n2m1;

		// MOLENAAR formulation
		Float T1 = 2.0*n3*(n2+2*n-1);
		Float B1=n2p1*(n4-1);

		Float T2=(n-1)*(3*n+1);
		Float B2=6*(n+1)*(n+1);

		Float T3=8.0*n4*(n4+1);
		Float B3=(n2+1)*(n4-1)*(n4-1);

		Float T4=n2*n2m12;
		Float B4=n2p12*n2p1;

		return (1.0/2.0-T1/B1+T2/B2+T3/B3*math::fastlog(n)-T4/B4*math::fastlog((n+1)/(n-1)));
	}

	Spectrum computeInterReflections(const Vector &H, const Float &phiH, const Vector &rVi, const Vector &rVo, const Spectrum &kd) const {
		if(Frame::cosTheta(rVo) <= 0.f || Frame::cosTheta(rVi) <= 0.f || Frame::cosTheta(H) <= 0.f)
			return Spectrum(0.f);

		Float ctiH = dot(H, rVi);
		if(ctiH <= 0.f)
			return Spectrum(0.f);

		Spectrum interGlobal(0.f);

		// === compute the V-cavity direction and the vector perpendicular to the V-cavity normal
		Vector vCavDir[2];
		vCavDir[0] = Vector(cos(phiH - M_PI_2), sin(phiH - M_PI_2), 0.0f);

		Vector xVectors[2];
		xVectors[0] = cross(vCavDir[0], H);
		xVectors[0] = normalize(xVectors[0]);
		xVectors[1] = Normal(-xVectors[0].x, -xVectors[0].y, xVectors[0].z);

		Vector Hvectors[2];
		Hvectors[0] = H;
		Hvectors[1] = Vector(-H.x, -H.y, H.z);

		vCavDir[1] = Vector(-vCavDir[0].x, -vCavDir[0].y, vCavDir[0].z);

		int nbSamples = m_nbInterSamples<=1?1:m_nbInterSamples;
		for(int i=0; i< nbSamples;i++) {

			int currentFacet = 0;
			Point pos = (Point)(xVectors[currentFacet] * (double) rand() / (RAND_MAX));

			Point newPos;
			if(!this->shadows(pos, rVi, Hvectors, xVectors, currentFacet)) { // === the chosen point is visible from rVi

				int nbLoops = 1; // === prevent to long light path
				Vector incidentDir = rVi;
				Spectrum throughput(1.0f);
				Vector outDir;

				int maxLoop = m_maxDepth<0?1000:m_maxDepth;
				while(nbLoops < maxLoop) {
					// === build the local frame
					Frame localFrame(xVectors[currentFacet], vCavDir[currentFacet], Hvectors[currentFacet]);
					// === choose a reflection direction
					throughput *= this->chooseDirection(incidentDir, localFrame, kd, outDir);

					// === Compute intersection with the micro-structure
					if(throughput.isZero())
						break; // === bad direction => break the while loop

					if(!this->intersects(pos, outDir, Hvectors, xVectors, currentFacet, newPos))
						break; // === no contribution => break the while loop

					if(this->validFrontFace(Hvectors[currentFacet], rVo) && !this->shadows(newPos, rVo, Hvectors, xVectors, currentFacet)) {
						// === compute the BSDF for the last bounce (no specular contribution here)
						Float frm=INV_PI*m_invNi*m_invNi;
						frm*=this->T01(-outDir,Hvectors[currentFacet])*this->T01(rVo,Hvectors[currentFacet]);
						Spectrum den = Spectrum(1.f) - kd*m_r10;
						Spectrum pathContrib((kd/den)*frm);
						pathContrib *=  dot(Hvectors[currentFacet], rVo);

						interGlobal += throughput * pathContrib;
					}

					// === Russian roulette
					Float q = std::min(throughput.max(), (Float) 0.95f);
					if ((double) rand() / (RAND_MAX) >= q)
						// === Russian roulette says STOP
						break; // === break the while loop
					else {
						// === init values for next ray
						throughput /= q;
						nbLoops++;
						pos = newPos;
						incidentDir = -outDir;
					}
				}
			}
		}

		Float divideCos = nbSamples * Frame::cosTheta(rVo) * Frame::cosTheta(rVi) *  Frame::cosTheta(H);
		if(divideCos <= 0)
			return Spectrum(0.f);
		else
			return interGlobal * ctiH /** ctoH*/ /divideCos;
	}

	// === first check =  front face or backface
	bool validFrontFace(const Vector &N, const Vector &d) const {
		if (dot(d, N) <= 0.0f)
			return false; // === back face

		return true;
	}

	// === d starts from p
	// === Computes the intersection with ray starting from p in direction of d and the geometry
	// === return false if no intersection and true otherwise
	// === if intersection, change the facet index
	bool intersects(const Point &p, const Vector &d, const Vector *Hnormals, const Vector *xVectors, int &facetIndex, Point &intersection) const {
		Ray r(Point(p.x, p.y, p.z), d, 1.0f);
		Normal S(r.o.x, r.o.y, r.o.z);
		int otherFacet = 1 - facetIndex; // === test only the other facet
		Float t = -dot(Hnormals[otherFacet], S)
				/ dot(Hnormals[otherFacet], r.d);
		if (t > 0.0f) {
			Point intersect = r(t);
			Normal pp(intersect.x, intersect.y, intersect.z);
			Float dotProd = dot(pp, xVectors[otherFacet]);
			if (dotProd < 0.0f || dotProd > 1.0f)
				return false;

			facetIndex = otherFacet;
			intersection = intersect;

			return true;
		}
		return false;
	}

	// === same as intersects function but without changing the facet index and without the intersection computation
	bool shadows(const Point &p, const Vector &d, const Vector *Hnormals, const Vector *xVectors, const int &facetIndex) const {
		int face = facetIndex;
		Point intersection;

		return this->intersects(p, d, Hnormals, xVectors, face, intersection);
	}

	// === return fr * cos(theta) / pdf
	// === incidentDir starts from p
	// === lightPath = true means the path is created from the light source
	Spectrum chooseDirection(const Vector &incidentDir, const Frame &localFrame,
			const Spectrum &kd, Vector &outDir) const {
		// === use Fresnel coefficient for specular sampling
		Spectrum coeffFresnel(this->FresnelNi(localFrame.n, incidentDir));
		Float specularWeight = coeffFresnel.average();

		Spectrum bsdfWeight(0.0f);
		// === first, choose specular or smooth part of the brdf
		Float R1 = (double) rand() / (RAND_MAX);
		if (R1 < specularWeight) {
			// === specular component
			outDir = reflect(incidentDir, localFrame.n);
			outDir = normalize(outDir);

			Float dotProd = dot(localFrame.n, outDir);
			if (dotProd <= 0)
				return bsdfWeight;

			bsdfWeight = coeffFresnel * dotProd / (specularWeight);
		} else {
			Point2 sample((double) rand() / (RAND_MAX),(double) rand() / (RAND_MAX));
			outDir = warp::squareToCosineHemisphere(sample);
			outDir = localFrame.toWorld(outDir);

			Float frm=INV_PI*m_invNi*m_invNi;
			frm*=this->T01(incidentDir,localFrame.n)*this->T01(outDir,localFrame.n);
			Spectrum den = Spectrum(1.f) - kd*m_r10;
			bsdfWeight = ((kd/den)*frm);

			if (!bsdfWeight.isZero()) {
				// === Why no cos(theta0) here? Because we have fr* cos(theta0) / pdf with pdf = cos(thetaO) / PI
				Float pdf = (1.0f - specularWeight) * INV_PI;
				if(pdf!=0)
					bsdfWeight /= pdf;
			}
		}

		return bsdfWeight;
	}

	//=================================================================
	// Fresnel coefficient
	//=================================================================
	Float FresnelNi(const Vector &N, const Vector &V) const	{
		Float c=dot(N,V);
		if(c<0.0) return 0.0;
		Float g=sqrt(m_ni*m_ni+c*c-1.0f);
		if(g==0.f)
			return 0.f;
		Float gmc=g-c, gpc=g+c;
		Float cgpc=c*gpc-1.0, cgmc=c*gmc+1.0;
		Float F=0.5*(gmc*gmc)/(gpc*gpc)*(1.0+(cgpc*cgpc)/(cgmc*cgmc));
		return F;
	}

};

/**
 * GLSL port of the IL Beckman TS shader.
 */
class ILSTDTS_exactShader : public Shader {
public:
	ILSTDTS_exactShader(Renderer *renderer, const Texture *diffuseReflectance,
			const Texture *alpha, const Float &ni) : Shader(renderer, EBSDFShader),
			m_diffuseReflectance(diffuseReflectance), m_alpha(alpha){
		m_diffuseReflectanceShader = renderer->registerShaderForResource(m_diffuseReflectance.get());
		m_alphaShader = renderer->registerShaderForResource(m_alpha.get());

		/* Compute the reflectance at perpendicular incidence */
		m_R0 = fresnelDielectricExt(1.0f, 1.f/ni);
	}

	bool isComplete() const {
		return m_diffuseReflectanceShader.get() != NULL &&
					   m_alphaShader.get() != NULL;
	}

	void putDependencies(std::vector<Shader *> &deps) {
		deps.push_back(m_diffuseReflectanceShader.get());
		deps.push_back(m_alphaShader.get());
	}

	void cleanup(Renderer *renderer) {
		renderer->unregisterShaderForResource(m_diffuseReflectance.get());
		renderer->unregisterShaderForResource(m_alpha.get());
	}

	void resolve(const GPUProgram *program, const std::string &evalName, std::vector<int> &parameterIDs) const {
		parameterIDs.push_back(program->getParameterID(evalName + "_R0", false));
	}

	void bind(GPUProgram *program, const std::vector<int> &parameterIDs, int &textureUnitOffset) const {
		program->setParameter(parameterIDs[0], m_R0);
	}

	void generateCode(std::ostringstream &oss,
			const std::string &evalName,
			const std::vector<std::string> &depNames) const {
		oss << "uniform float " << evalName << "_R0;" << endl
			<< endl
			<< "float " << evalName << "_D(vec3 m, float alpha) {" << endl
			<< "    float ct = cosTheta(m);" << endl
			<< "    if (cosTheta(m) <= 0.0)" << endl
			<< "        return 0.0;" << endl
			<< "    float ex = tanTheta(m) / alpha;" << endl
			<< "    return exp(-(ex*ex)) / (pi * alpha * alpha *" << endl
			<< "               pow(cosTheta(m), 4.0));" << endl
			<< "}" << endl
			<< endl
			<< "float " << evalName << "_G(vec3 m, vec3 wi, vec3 wo) {" << endl
			<< "    if ((dot(wi, m) * cosTheta(wi)) <= 0 || " << endl
			<< "        (dot(wo, m) * cosTheta(wo)) <= 0)" << endl
			<< "        return 0.0;" << endl
			<< "    float nDotM = cosTheta(m);" << endl
			<< "    return min(1.0, min(" << endl
			<< "        abs(2 * nDotM * cosTheta(wo) / dot(wo, m))," << endl
			<< "        abs(2 * nDotM * cosTheta(wi) / dot(wi, m))));" << endl
			<< "}" << endl
			<< endl
			<< endl
			<< "float " << evalName << "_schlick(float ct) {" << endl
			<< "    float ctSqr = ct*ct, ct5 = ctSqr*ctSqr*ct;" << endl
			<< "    return " << evalName << "_R0 + (1.0 - " << evalName << "_R0) * ct5;" << endl
			<< "}" << endl
			<< endl
			<< "vec3 " << evalName << "(vec2 uv, vec3 wi, vec3 wo) {" << endl
			<< "    if (cosTheta(wi) <= 0 || cosTheta(wo) <= 0)" << endl
			<< "        return vec3(0.0);" << endl
			<< "    vec3 H = normalize(wi + wo);" << endl
			<< "    vec3 diffuseRef = " << depNames[0] << "(uv);" << endl
			<< "    float alpha = max(0.2, " << depNames[1] << "(uv)[0]);" << endl
			<< "    float D = " << evalName << "_D(H, alpha)" << ";" << endl
			<< "    float G = " << evalName << "_G(H, wi, wo);" << endl
			<< "    float F = " << evalName << "_schlick(1-dot(wi, H));" << endl
			<< "    return vec3(1.0) * (F * D * G / (4*cosTheta(wi))) + " << endl
			<< "           diffuseRef * ((1-F) * cosTheta(wo) * inv_pi);" << endl
			<< "}" << endl
			<< endl
			<< "vec3 " << evalName << "_diffuse(vec2 uv, vec3 wi, vec3 wo) {" << endl
			<< "    if (cosTheta(wi) < 0.0 || cosTheta(wo) < 0.0)" << endl
			<< "    	return vec3(0.0);" << endl
			<< "    vec3 diffuseRef = " << depNames[0] << "(uv);" << endl
			<< "    return diffuseRef * inv_pi * cosTheta(wo);"<< endl
			<< "}" << endl;
	}
	MTS_DECLARE_CLASS()
private:
	ref<const Texture> m_diffuseReflectance;
	ref<const Texture> m_alpha;
	ref<Shader> m_diffuseReflectanceShader;
	ref<Shader> m_alphaShader;
	ref<Shader> m_niShader;
	Float m_R0;
};

Shader *ILSTD_exact::createShader(Renderer *renderer) const {
	return new ILSTDTS_exactShader(renderer,
			m_diffuseReflectance.get(), new ConstantFloatTexture(m_alpha), m_ni);
}

MTS_IMPLEMENT_CLASS(ILSTDTS_exactShader, false, Shader)
MTS_IMPLEMENT_CLASS_S(ILSTD_exact, false, BSDF)
MTS_EXPORT_PLUGIN(ILSTD_exact, "IL exact model with approximate STD distribution and Torrance-Sparrow GAF");
MTS_NAMESPACE_END





