/*
 * microfacetSTD.h
 *
 *  Created on: 9 déc. 2016
 *      Author: Mickael Ribardiere
 */

#ifndef SRC_BSDFS_MICROFACETSTD_H_
#define SRC_BSDFS_MICROFACETSTD_H_



#include <mitsuba/mitsuba.h>
#include <mitsuba/core/frame.h>
#include <mitsuba/core/properties.h>
#include <boost/algorithm/string.hpp>

#include <gsl/gsl_errno.h>
#include <gsl/gsl_sf_gamma.h>
#include <gsl/gsl_sf_hyperg.h>

#include "STDtools.h"


MTS_NAMESPACE_BEGIN

/**
 * \brief Implementation of some STD family distributions: t student and Hyper-Cauchy
 * distributions + various useful sampling routines.
 *
 * The code is directly inspired from the original microfacet.h file in mitsuba
 *
 */


class MicrofacetDistributionSTD {
public:
	/// Supported distribution types
	enum ETypeSTD {
		// === STD family distributions
		EStudentT            	= 0,
		EStudentT_discrete      = 1,
		EStudentT_halfDiscrete	= 2,
		EStudentT_approx		= 3,	// === Used the approximated lambda function
		EHyperCauchy			= 4
	};

	/**
	 * Create an isotropic microfacet distribution of the specified type
	 *
	 * \param type
	 *     The desired type of microfacet distribution
	 * \param alpha
	 *     The surface roughness
	 */
	inline MicrofacetDistributionSTD(ETypeSTD type, Float alpha, bool sampleVisible = true, Float gamma = 2.0, bool correlated = false)
		: m_type(type), m_alphaU(alpha), m_alphaV(alpha), m_sampleVisible(sampleVisible), m_correlatedGAF(correlated) {
		m_alphaU = std::max(m_alphaU, (Float) 1e-4f);
		m_alphaV = std::max(m_alphaV, (Float) 1e-4f);

		m_gamma = gamma;
		computeGammaExponent();
	}

	/**
	 * Create an anisotropic microfacet distribution of the specified type
	 *
	 * \param type
	 *     The desired type of microfacet distribution
	 * \param alphaU
	 *     The surface roughness in the tangent direction
	 * \param alphaV
	 *     The surface roughness in the bitangent direction
	 */
	inline MicrofacetDistributionSTD(ETypeSTD type, Float alphaU, Float alphaV, bool sampleVisible = true, Float gamma = 2.0, bool correlated = false)
		: m_type(type), m_alphaU(alphaU), m_alphaV(alphaV), m_sampleVisible(sampleVisible), m_correlatedGAF(correlated) {
		m_alphaU = std::max(m_alphaU, (Float) 1e-4f);
		m_alphaV = std::max(m_alphaV, (Float) 1e-4f);

		m_gamma = gamma;
		computeGammaExponent();
	}

	/**
	 * \brief Create a microfacet distribution from a Property data
	 * structure
	 */
	MicrofacetDistributionSTD(const Properties &props, ETypeSTD type = EStudentT,
		Float alphaU = 0.1f, Float alphaV = 0.1f, bool sampleVisible = true, Float gamma = 2.0, bool correlated = false)
		: m_type(type), m_alphaU(alphaU), m_alphaV(alphaV) {

		if (props.hasProperty("distribution")) {
			std::string distr = boost::to_lower_copy(props.getString("distribution"));
			if (distr == "studentt_d")
				m_type = EStudentT_discrete;
			else if (distr == "studentt")
				m_type = EStudentT;
			else if (distr == "studentt_hd")
				m_type = EStudentT_halfDiscrete;
			else if (distr == "studentt_approx")
				m_type = EStudentT_approx;
			else if (distr == "hypercauchy")
				m_type = EHyperCauchy;
			else
				SLog(EError, "Specified an invalid distribution \"%s\", must be "
					"\"studentt_d\", \"studentt\", \"studentt_hd\", \"studentt_approx\", \"hypercauchy\"!", distr.c_str());
		}

		if (props.hasProperty("alpha")) {
			m_alphaU = m_alphaV = props.getFloat("alpha");
			if (props.hasProperty("alphaU") || props.hasProperty("alphaV"))
				SLog(EError, "Microfacet model: please specify either 'alpha' or 'alphaU'/'alphaV'.");
		} else if (props.hasProperty("alphaU") || props.hasProperty("alphaV")) {
			if (!props.hasProperty("alphaU") || !props.hasProperty("alphaV"))
				SLog(EError, "Microfacet model: both 'alphaU' and 'alphaV' must be specified.");
			if (props.hasProperty("alpha"))
				SLog(EError, "Microfacet model: please specify either 'alpha' or 'alphaU'/'alphaV'.");
			m_alphaU = props.getFloat("alphaU");
			m_alphaV = props.getFloat("alphaV");
		}

		if (m_alphaU == 0 || m_alphaV == 0) {
			SLog(EWarn, "Cannot create a microfacet distribution with alphaU/alphaV=0 (clamped to 0.0001)."
					"Please use the corresponding smooth reflectance model to get zero roughness.");
		}

		m_alphaU = std::max(m_alphaU, (Float) 1e-4f);
		m_alphaV = std::max(m_alphaV, (Float) 1e-4f);

		m_sampleVisible = props.getBoolean("sampleVisible", sampleVisible);
		m_correlatedGAF = props.getBoolean("correlated", correlated);

		m_gamma = props.getFloat("gamma",gamma);
		computeGammaExponent();
	}

	/// Return the distribution type
	inline ETypeSTD getType() const { return m_type; }

	/// Return the roughness (isotropic case)
	inline Float getAlpha() const { return m_alphaU; }

	/// Return the roughness along the tangent direction
	inline Float getAlphaU() const { return m_alphaU; }

	/// Return the roughness along the bitangent direction
	inline Float getAlphaV() const { return m_alphaV; }

	/// Return the gamma component (for the PGTR distribution)
	inline Float getGamma() const { return m_gamma; }

	/// Scale the roughness values by some constant
	inline void scaleGamma(Float value) {
		m_gamma *= value;
		computeGammaExponent();
	}

	/// Return whether or not only visible normals are sampled?
	inline bool getSampleVisible() const { return m_sampleVisible; }
	/// Return whether or not the GAF is correlated?
	inline bool getCorrelatedGAF() const { return m_correlatedGAF; }

	/// Is this an anisotropic microfacet distribution?
	inline bool isAnisotropic() const { return m_alphaU != m_alphaV; }

	/// Is this an anisotropic microfacet distribution?
	inline bool isIsotropic() const { return m_alphaU == m_alphaV; }

	/// Scale the roughness values by some constant
	inline void scaleAlpha(Float value) {
		m_alphaU *= value;
		m_alphaV *= value;
	}

	/**
	 * \brief Evaluate the microfacet distribution function
	 *
	 * \param m
	 *     The microfacet normal
	 */
	inline Float eval(const Vector &m) const {
		if (Frame::cosTheta(m) <= 0)
			return 0.0f;

		Float cosTheta2 = Frame::cosTheta2(m);
		Float beckmannExponent = ((m.x*m.x) / (m_alphaU * m_alphaU)
				+ (m.y*m.y) / (m_alphaV * m_alphaV)) / cosTheta2;

		Float result = 0;

		switch (m_type) {
			case EStudentT:
			case EStudentT_halfDiscrete:
			case EStudentT_approx:
			case EStudentT_discrete: {
				Float root = (1.0f + beckmannExponent/m_gammaMinOne);
				result = (Float) 1 / (M_PI * m_alphaU * m_alphaV * cosTheta2 * cosTheta2*  pow(root,m_gamma));
			}
			break;

			case EHyperCauchy: {
				Float root = (1.0f + beckmannExponent);
				result = (Float) (m_gamma-1) / (M_PI * m_alphaU * m_alphaV * cosTheta2 * cosTheta2*  pow(root,m_gamma));
			}
			break;

		default:
			SLog(EError, "Invalid distribution type!");
			return -1;
		}
		/* Prevent potential numerical issues in other stages of the model */
		if (result * Frame::cosTheta(m) < 1e-20f)
			result = 0;

		return result;
	}

	/**
	 * \brief Wrapper function which calls \ref sampleAll() or \ref sampleVisible()
	 * depending on the parameters of this class
	 */
	inline Normal sample(const Vector &wi, const Point2 &sample, Float &pdf) const {
		Normal m;
		if (m_sampleVisible && m_type != EHyperCauchy) {
			m = sampleVisible(wi, sample);
			pdf = pdfVisible(wi, m);
		} else {
			m = sampleAll(sample, pdf);
		}
		return m;
	}

	/**
	 * \brief Wrapper function which calls \ref sampleAll() or \ref sampleVisible()
	 * depending on the parameters of this class
	 */
	inline Normal sample(const Vector &wi, const Point2 &sample) const {
		Normal m;
		if (m_sampleVisible && m_type != EHyperCauchy) {
			m = sampleVisible(wi, sample);
		} else {
			Float pdf;
			m = sampleAll(sample, pdf);
		}
		return m;
	}

	/**
	 * \brief Wrapper function which calls \ref pdfAll() or \ref pdfVisible()
	 * depending on the parameters of this class
	 */
	inline Float pdf(const Vector &wi, const Vector &m) const {
		if (m_sampleVisible && m_type != EHyperCauchy)
			return pdfVisible(wi, m);
		else
			return pdfAll(m);
	}

	/**
	 * \brief Draw a sample from the microfacet normal distribution
	 * (including *all* normals) and return the associated
	 * probability density
	 *
	 * \param sample
	 *    A uniformly distributed 2D sample
	 * \param pdf
	 *    The probability density wrt. solid angles
	 */
	inline Normal sampleAll(const Point2 &sample, Float &pdf) const {
		/* The azimuthal component is always selected
		   uniformly regardless of the distribution */
		Float cosThetaM = 0.0f;
		Float sinPhiM, cosPhiM;
		Float alphaSqr = 0.f;

		switch (m_type) {
			case EStudentT:
			case EStudentT_halfDiscrete:
			case EStudentT_approx:
			case EStudentT_discrete: {
					if (isIsotropic()) {
						/* Sample phi component (isotropic case) */
						math::sincos((2.0f * M_PI) * sample.y, &sinPhiM, &cosPhiM);

						/* Sample theta component */
						alphaSqr = m_gammaMinOne*m_alphaU*m_alphaU;
					} else {
						/* Sample phi component (anisotropic case) */
						Float phiM = std::atan(m_alphaV / m_alphaU *
							std::tan(M_PI + 2*M_PI*sample.y)) + M_PI * std::floor(2*sample.y + 0.5f);
						math::sincos(phiM, &sinPhiM, &cosPhiM);

						Float cosSc = cosPhiM / m_alphaU, sinSc = sinPhiM / m_alphaV;
						alphaSqr = m_gammaMinOne / (cosSc*cosSc + sinSc*sinSc);
					}

					/* Sample theta component */
					Float power = 1.0f/(1.0f-m_gamma);
					Float tanThetaMSqr = alphaSqr * (pow(1.0f-sample.x,power) - 1.0f);
					cosThetaM = 1.0f / std::sqrt(1.0f + tanThetaMSqr);

					/* Compute probability density of the sampled position */
					Float temp = 1+tanThetaMSqr/alphaSqr;
					pdf = INV_PI / (m_alphaU*m_alphaV*cosThetaM*cosThetaM*cosThetaM*pow(temp,m_gamma));
				}
				break;

			case EHyperCauchy: {
				if (isIsotropic()) {
					/* Sample phi component (isotropic case) */
					math::sincos((2.0f * M_PI) * sample.y, &sinPhiM, &cosPhiM);

					/* Sample theta component */
					alphaSqr = m_alphaU*m_alphaU;
				} else {
					/* Sample phi component (anisotropic case) */
					Float phiM = std::atan(m_alphaV / m_alphaU *
						std::tan(M_PI + 2*M_PI*sample.y)) + M_PI * std::floor(2*sample.y + 0.5f);
					math::sincos(phiM, &sinPhiM, &cosPhiM);

					Float cosSc = cosPhiM / m_alphaU, sinSc = sinPhiM / m_alphaV;
					alphaSqr = 1.0 / (cosSc*cosSc + sinSc*sinSc);
				}

				/* Sample theta component */
				Float power = 1.0f/(1.0f-m_gamma);
				Float tanThetaMSqr = alphaSqr * (pow(1.0f-sample.x,power) - 1.0f);
				cosThetaM = 1.0f / std::sqrt(1.0f + tanThetaMSqr);

				/* Compute probability density of the sampled position */
				Float temp = 1+tanThetaMSqr/alphaSqr;
				pdf = (m_gamma-1.0) * INV_PI / (m_alphaU*m_alphaV*cosThetaM*cosThetaM*cosThetaM*pow(temp,m_gamma));
			}
			break;

			default:
				SLog(EError, "Invalid distribution type!");
				pdf = -1;
				return Vector(-1);
		}

		/* Prevent potential numerical issues in other stages of the model */
		if (pdf < 1e-20f)
			pdf = 0;

		Float sinThetaM = std::sqrt(
			std::max((Float) 0, 1 - cosThetaM*cosThetaM));

		return Vector(
			sinThetaM * cosPhiM,
			sinThetaM * sinPhiM,
			cosThetaM
		);
	}

	/**
	 * \brief Returns the density function associated with
	 * the \ref sampleAll() function.
	 *
	 * \param m
	 *     The microfacet normal
	 */
	inline Float pdfAll(const Vector &m) const {
		/* PDF is just D(m) * cos(theta_M) */
		return eval(m) * Frame::cosTheta(m);
	}


	/**
	 * \brief Draw a sample from the distribution of visible normals
	 * and return the associated probability density
	 *
	 * \param _wi
	 *    A reference direction that defines the set of visible normals
	 * \param sample
	 *    A uniformly distributed 2D sample
	 * \param pdf
	 *    The probability density wrt. solid angles
	 */
	inline Normal sampleVisible(const Vector &_wi, const Point2 &sample) const {

		bool beckmann = true;
		if(m_alphaU < 0.01 || m_alphaV < 0.01) {
			Float pdf;
			return sampleAll(sample, pdf);
		} else if (m_gamma > 10)
			beckmann = true;
		else {
			Float minAlpha = std::min(m_alphaU,m_alphaV);

			Float GGXupperBound = std::min(5.0, 2.0+minAlpha/0.1);
			Float BecklowerBound = std::max(5.0, 10.0-5.0* std::max(0.0,minAlpha-0.2)/0.1 );

			if(m_gamma < GGXupperBound)
				beckmann = false;
			else if(m_gamma > BecklowerBound)
				beckmann = true;
			else {
				Float pdf;
				return sampleAll(sample, pdf);
			}
		}

		/* Step 1: stretch wi */
		Vector wi = normalize(Vector(
			m_alphaU * _wi.x,
			m_alphaV * _wi.y,
			_wi.z
		));

		/* Get polar coordinates */
		Float theta = 0, phi = 0;
		if (wi.z < (Float) 0.99999) {
			theta = std::acos(wi.z);
			phi = std::atan2(wi.y, wi.x);
		}
		Float sinPhi, cosPhi;
		math::sincos(phi, &sinPhi, &cosPhi);

		/* Step 2: simulate P22_{wi}(slope.x, slope.y, 1, 1) */
		Vector2 slope = sampleVisible11(theta, sample, beckmann);

		/* Step 3: rotate */
		slope = Vector2(
			cosPhi * slope.x - sinPhi * slope.y,
			sinPhi * slope.x + cosPhi * slope.y);

		/* Step 4: unstretch */
		slope.x *= m_alphaU;
		slope.y *= m_alphaV;

		/* Step 5: compute normal */
		Float normalization = (Float) 1 / std::sqrt(slope.x*slope.x
				+ slope.y*slope.y + (Float) 1.0);

		return Normal(
			-slope.x * normalization,
			-slope.y * normalization,
			normalization
		);
	}

	/// Implements the probability density of the function \ref sampleVisible()
	Float pdfVisible(const Vector &wi, const Vector &m) const {
		if(Frame::cosTheta(wi) == 0)
			return 0.0f;

		if(m_alphaU < 0.01 || m_alphaV < 0.01)
			return pdfAll(m);
		else if (m_gamma > 10)
			return smithG1Beckmann(wi, m) * absDot(wi, m) * DBeck(m) / std::abs(Frame::cosTheta(wi));
		else {
			Float minAlpha = std::min(m_alphaU,m_alphaV);

			Float GGXupperBound = std::min(5.0, 2.0+minAlpha/0.1);
			Float BecklowerBound = std::max(5.0, 10.0-5.0* std::max(0.0,minAlpha-0.2)/0.1 );

			if(m_gamma < GGXupperBound)
				return smithG1GGX(wi, m) * absDot(wi, m) * DGGX(m) / std::abs(Frame::cosTheta(wi));
			else if(m_gamma > BecklowerBound)
				return smithG1Beckmann(wi, m) * absDot(wi, m) * DBeck(m) / std::abs(Frame::cosTheta(wi));
			else
				return pdfAll(m);
		}


		//return smithG1(wi, m) * absDot(wi, m) * eval(m) / std::abs(Frame::cosTheta(wi));
	}

	double smithG1(const Vector &v, const Vector &m) const {
		/* Ensure consistent orientation (can't see the back
		   of the microfacet from the front and vice versa) */
		if (dot(v, m) * Frame::cosTheta(v) <= 0)
			return 0.0f;

		return std::min(1.0 / (1.0 + lambda(v)),1.0);
	}

	// === the lambda function
	double lambda(const Vector &v) const {
		/* Perpendicular incidence -- no shadowing/masking */
		Float tanTheta = std::abs(Frame::tanTheta(v));
		if (tanTheta == 0.0f)
			return 0.0f;

		Float alpha = projectRoughness(v);
		double lambda = 0.0;
		switch (m_type) {
			case EStudentT_halfDiscrete: {
					// === half integer discrete version
					double root = alpha * tanTheta;
					double rootInv = 1.0 / root;
					double rootInvSqr = rootInv*rootInv;

					double A2 = 0.f;

					int gIndex = (int)m_gamma - 2;

					double A1 = half_integerA1[gIndex];

					double valA2 = std::atan(rootInv/m_sqrtGammaMinOne);
					A2 = half_integerA23[gIndex]*(M_PI_2-valA2);
					Float gammaModif =  2*m_gamma-3;
					for(int k=0; k<= gIndex;k++)
						A2 -= half_integerA22[gIndex][k]*(FastSin((gammaModif-2*k)*valA2)); // === fast sin version


					lambda = A1*(m_sqrtGammaMinOne*root*pow(1.f+rootInvSqr/m_gammaMinOne,1.5-m_gamma)/(2.0*m_gamma-3.0)-A2);
				}
				break;

			case EStudentT_discrete: {
					// === integer discrete version
					double root = alpha * tanTheta;
					double rootInv = 1.0 / root;
					double rootInvSqr = rootInv*rootInv;

					double A2 = 0.f;

					int gIndex = (int)m_gamma - 2;

					double A1 = integerA1[gIndex];

					double valA2 = rootInv/(sqrt(m_gammaMinOne+rootInvSqr));
					double power = valA2;
					valA2 *= valA2;
					for(int k=0; k<= gIndex;k++) {
						A2 += integerA2[gIndex][k]*(1.0-power);
						power *= valA2;
					}

					lambda = A1*(m_sqrtGammaMinOne*root*pow(1.f+rootInvSqr/m_gammaMinOne,1.5-m_gamma)/(2.0*m_gamma-3.0)-A2);
				}
				break;


			case EStudentT_approx: {
					double root = alpha * tanTheta;
					double rootInv = 1.0 / root;
					double rootInvSqr = rootInv*rootInv;

					double S1 = root * pow(m_gammaMinOne+rootInvSqr,1.5-m_gamma);

					double rootInv3 = rootInvSqr*rootInv;

					double S2 = (1.066*rootInv + 2.655*rootInvSqr + 4.892*rootInv3)/(1.038 + 2.969*rootInv + 4.305*rootInvSqr + 4.418*rootInv3);
					double factor =  (6.537 + 6.074*rootInv - 0.623*rootInvSqr + 5.223*rootInv3)/(6.538 + 6.103* rootInv - 3.218*rootInvSqr + 6.347*rootInv3);

					S2 *= m_S2_G1 + m_S2_G2 * factor;

					S1 *= m_F1;
					S2 *= m_sqrtGammaMinOne;

					lambda = (S1+S2)*m_F2 - 0.5f;
				}
				break;

			case EStudentT: {
					double root = alpha * tanTheta;
					double rootInv = 1.0 / root;
					double rootInvSqr = rootInv*rootInv;

					double S1 = root * pow(m_gammaMinOne+rootInvSqr,1.5-m_gamma);

					double hypVal = -rootInvSqr/m_gammaMinOne;
					double S2 = rootInv * this->hypergeometricFunction(hypVal);

					S1 *= m_F1;
					S2 *= m_sqrtGammaMinOne;

					lambda = (S1+S2)*m_F2 - 0.5f;
				}
				break;

			case EHyperCauchy: {
					double root = alpha * tanTheta;
					double rootInv = 1.0 / root;
					double rootInvSqr = rootInv*rootInv;

					double S1 = root * pow(1.0+rootInvSqr,1.5-m_gamma);

					double hypVal = -rootInvSqr;
					double S2 = rootInv * this->hypergeometricFunction(hypVal);

					S1 *= m_F1;
					S2 *= m_gammaMinOne;

					lambda = (S1+S2)*m_F2 - 0.5f;
				}
				break;

			default:
				SLog(EError, "Invalid distribution type!");
				return -1.0f;
		}
		return std::max(lambda,0.0);
	}

	/**
	 * \brief Separable shadow-masking function based on Smith's
	 * one-dimensional masking model
	 */
	inline Float G(const Vector &wi, const Vector &wo, const Vector &m) const {
		if(m_correlatedGAF) {
			if (dot(wi, m) * Frame::cosTheta(wi) <= 0)
				return 0.0f;
			if (dot(wo, m) * Frame::cosTheta(wo) <= 0)
				return 0.0f;
			return std::min(1.0/(1.0+lambda(wi)+lambda(wo)),1.0);
		} else
			return  smithG1(wi, m)*smithG1(wo, m);
	}

	/// Return a string representation of the name of a distribution
	inline static std::string distributionName(ETypeSTD type) {
		switch (type) {
			case EStudentT_discrete: return "student t discrete"; break;
			case EStudentT: return "student t"; break;
			case EStudentT_halfDiscrete: return "student t half discrete"; break;
			case EStudentT_approx: return "student t approx"; break;
			default: return "invalid"; break;
		}
	}

	/// Return a string representation of the contents of this instance
	std::string toString() const {
		return formatString("MicrofacetDistributionSTD[type=\"%s\", alphaU=%f, alphaV=%f]",
			distributionName(m_type).c_str(), m_alphaU, m_alphaV);
	}
protected:
	/// Compute the effective roughness projected on direction \c v
	inline Float projectRoughness(const Vector &v) const {
		Float invSinTheta2 = 1 / Frame::sinTheta2(v);

		if (isIsotropic() || invSinTheta2 <= 0)
			return m_alphaU;

		Float cosPhi2 = v.x * v.x * invSinTheta2;
		Float sinPhi2 = v.y * v.y * invSinTheta2;

		return std::sqrt(cosPhi2 * m_alphaU * m_alphaU + sinPhi2 * m_alphaV * m_alphaV);
	}


	Float smithG1GGX(const Vector &v, const Vector &m) const {
		if (dot(v, m) * Frame::cosTheta(v) <= 0)
			return 0.0f;

		/* Perpendicular incidence -- no shadowing/masking */
		Float tanTheta = std::abs(Frame::tanTheta(v));
		if (tanTheta == 0.0f)
			return 1.0f;

		Float alpha = projectRoughness(v);

		Float root = alpha * tanTheta;
		return 2.0f / (1.0f + math::hypot2((Float) 1.0f, root));
	}

	inline Float DGGX(const Vector &m) const {
		if (Frame::cosTheta(m) <= 0)
			return 0.0f;

		Float cosTheta2 = Frame::cosTheta2(m);
		Float beckmannExponent = ((m.x*m.x) / (m_alphaU * m_alphaU)
				+ (m.y*m.y) / (m_alphaV * m_alphaV)) / cosTheta2;

		Float root = ((Float) 1 + beckmannExponent) * cosTheta2;
		Float result = (Float) 1 / (M_PI * m_alphaU * m_alphaV * root * root);

		if (result * Frame::cosTheta(m) < 1e-20f)
			result = 0;

		return result;
	}

	Float smithG1Beckmann(const Vector &v, const Vector &m) const {
		if (dot(v, m) * Frame::cosTheta(v) <= 0)
			return 0.0f;

		/* Perpendicular incidence -- no shadowing/masking */
		Float tanTheta = std::abs(Frame::tanTheta(v));
		if (tanTheta == 0.0f)
			return 1.0f;

		Float alpha = projectRoughness(v);

		Float a = 1.0f / (alpha * tanTheta);
		if (a >= 1.6f)
			return 1.0f;

		Float aSqr = a*a;
		return (3.535f * a + 2.181f * aSqr)
			 / (1.0f + 2.276f * a + 2.577f * aSqr);
	}

	inline Float DBeck(const Vector &m) const {
		if (Frame::cosTheta(m) <= 0)
			return 0.0f;

		Float cosTheta2 = Frame::cosTheta2(m);
		Float beckmannExponent = ((m.x*m.x) / (m_alphaU * m_alphaU)
				+ (m.y*m.y) / (m_alphaV * m_alphaV)) / cosTheta2;

		Float result = math::fastexp(-beckmannExponent) /
			(M_PI * m_alphaU * m_alphaV * cosTheta2 * cosTheta2);

		if (result * Frame::cosTheta(m) < 1e-20f)
			result = 0;

		return result;
	}

	/**
	 * \brief Visible normal sampling code for the alpha=1 case
	 *
	 * Source: supplemental material of "Importance Sampling
	 * Microfacet-Based BSDFs using the Distribution of Visible Normals"
	 */
	Vector2 sampleVisible11(Float thetaI, Point2 sample, bool beckmann) const {
		const Float SQRT_PI_INV = 1 / std::sqrt(M_PI);
		Vector2 slope;

		if(beckmann) { // === use Beckmann
				/* Special case (normal incidence) */
				if (thetaI < 1e-4f) {
					Float sinPhi, cosPhi;
					Float r = std::sqrt(-math::fastlog(1.0f-sample.x));
					math::sincos(2 * M_PI * sample.y, &sinPhi, &cosPhi);
					return Vector2(r * cosPhi, r * sinPhi);
				}

				/* The original inversion routine from the paper contained
				   discontinuities, which causes issues for QMC integration
				   and techniques like Kelemen-style MLT. The following code
				   performs a numerical inversion with better behavior */
				Float tanThetaI = std::tan(thetaI);
				Float cotThetaI = 1 / tanThetaI;

				/* Search interval -- everything is parameterized
				   in the erf() domain */
				Float a = -1, c = math::erf(cotThetaI);
				Float sample_x = std::max(sample.x, (Float) 1e-6f);

				/* Start with a good initial guess */
				//Float b = (1-sample_x) * a + sample_x * c;

				/* We can do better (inverse of an approximation computed in Mathematica) */
				Float fit = 1 + thetaI*(-0.876f + thetaI * (0.4265f - 0.0594f*thetaI));
				Float b = c - (1+c) * std::pow(1-sample_x, fit);

				/* Normalization factor for the CDF */
				Float normalization = 1 / (1 + c + SQRT_PI_INV*
					tanThetaI*std::exp(-cotThetaI*cotThetaI));

				int it = 0;
				while (++it < 10) {
					/* Bisection criterion -- the oddly-looking
					   boolean expression are intentional to check
					   for NaNs at little additional cost */
					if (!(b >= a && b <= c))
						b = 0.5f * (a + c);

					/* Evaluate the CDF and its derivative
					   (i.e. the density function) */
					Float invErf = math::erfinv(b);
					Float value = normalization*(1 + b + SQRT_PI_INV*
						tanThetaI*std::exp(-invErf*invErf)) - sample_x;
					Float derivative = normalization * (1
						- invErf*tanThetaI);

					if (std::abs(value) < 1e-5f)
						break;

					/* Update bisection intervals */
					if (value > 0)
						c = b;
					else
						a = b;

					b -= value / derivative;
				}

				/* Now convert back into a slope value */
				slope.x = math::erfinv(b);

				/* Simulate Y component */
				slope.y = math::erfinv(2.0f*std::max(sample.y, (Float) 1e-6f) - 1.0f);
		} else { // === use GGX
				/* Special case (normal incidence) */
				if (thetaI < 1e-4f) {
					Float sinPhi, cosPhi;
					Float r = math::safe_sqrt(sample.x / (1 - sample.x));
					math::sincos(2 * M_PI * sample.y, &sinPhi, &cosPhi);
					return Vector2(r * cosPhi, r * sinPhi);
				}

				/* Precomputations */
				Float tanThetaI = std::tan(thetaI);
				Float a = 1 / tanThetaI;
				Float G1 = 2.0f / (1.0f + math::safe_sqrt(1.0f + 1.0f / (a*a)));

				/* Simulate X component */
				Float A = 2.0f * sample.x / G1 - 1.0f;
				if (std::abs(A) == 1)
					A -= math::signum(A)*Epsilon;
				Float tmp = 1.0f / (A*A - 1.0f);
				Float B = tanThetaI;
				Float D = math::safe_sqrt(B*B*tmp*tmp - (A*A - B*B) * tmp);
				Float slope_x_1 = B * tmp - D;
				Float slope_x_2 = B * tmp + D;
				slope.x = (A < 0.0f || slope_x_2 > 1.0f / tanThetaI) ? slope_x_1 : slope_x_2;

				/* Simulate Y component */
				Float S;
				if (sample.y > 0.5f) {
					S = 1.0f;
					sample.y = 2.0f * (sample.y - 0.5f);
				} else {
					S = -1.0f;
					sample.y = 2.0f * (0.5f - sample.y);
				}

				/* Improved fit */
				Float z =
					(sample.y * (sample.y * (sample.y * (-(Float) 0.365728915865723) + (Float) 0.790235037209296) -
						(Float) 0.424965825137544) + (Float) 0.000152998850436920) /
					(sample.y * (sample.y * (sample.y * (sample.y * (Float) 0.169507819808272 - (Float) 0.397203533833404) -
						(Float) 0.232500544458471) + (Float) 1) - (Float) 0.539825872510702);

				slope.y = S * z * std::sqrt(1.0f + slope.x*slope.x);
		}
		return slope;
	}

	/// Helper routine: convert from Beckmann-style roughness values to Phong exponents (Walter et al.)
	void computeGammaExponent() {
		m_gamma = std::max(m_gamma, 1.51f);
		if(m_type == EStudentT_discrete) {
			m_gamma = std::min(m_gamma, 50.f);
			m_gamma = (int)m_gamma;
		}else if(m_type == EStudentT_halfDiscrete) {
			m_gamma = std::min(m_gamma, 50.f);
			m_gamma = (int)m_gamma + 0.5;
		} else
			gsl_set_error_handler_off(); // === GSL is only used for the real version
		m_gammaMinOne = m_gamma-1.f;
		m_sqrtGammaMinOne = sqrt(m_gammaMinOne);
		if(m_type == EHyperCauchy)
			m_F1 = m_gammaMinOne/(2.0f*m_gamma-3.0f);
		else
			m_F1 = pow(m_gammaMinOne,m_gamma)/(2.0f*m_gamma-3.0f);
		m_F2 = gammaApp(m_gamma-0.5f)/(gammaApp(m_gamma)*s_sqrtPi);

		if(m_type == EStudentT_approx) {
			double gsquared = m_gamma*m_gamma;
			double g3 = gsquared*m_gamma;
			m_S2_G1 = (14.402 - 27.145*m_gamma + 20.574*gsquared - 2.745*g3)/
					(-30.612 + 86.567*m_gamma - 84.341*gsquared + 29.938*g3);
			m_S2_G2 = (-129.404 + 324.987*m_gamma - 299.305*gsquared + 93.268*g3) /
					(-92.609 + 256.006*m_gamma - 245.663*gsquared + 86.064*g3);
		} else {
			m_S2_G1 = 0;
			m_S2_G2 = 0;
		}
	}

	// === 2F1 function for our distribution function
	Float hypergeometricFunction(double z) const {
		if (z== 0.f)
			return std::numeric_limits<Float>::infinity();

		Float a = 0.5f;
		Float b = m_gamma-0.5f;
		Float c = 1.5f;
		if(z <= -1.0f) { // === GSL hyperg_2F1 is defined for -1 < z <= 1
			double normalZ = z/(z-1.0f);

			if (normalZ == 1.f)	// === when z tends to - Infinity
				return 0.f;
			if (m_gamma >= 12 && normalZ >= 0.99f) {
				// === Abremowitz and Stegun 15.3.7 (for gamma not integer)
				double gammaA = 1.77245f;
				double gammaB = gammaApp(b);
				double gammaC = 0.886227f;
				double invZ = 1.f/z;

				double part1 = gammaC*gammaApp(b-a)/(gammaB);
				part1 *= pow(-z,-a)*gsl_sf_hyperg_2F1(a,1-c+a,1-b+a,invZ);

				double part2 = gammaC*gsl_sf_gamma(a-b)/(gammaA*gsl_sf_gamma(c-b));
				part2 *= pow(-z,-b)*gsl_sf_hyperg_2F1(b,1-c+b,1-a+b,invZ);

				return part1+part2;
			} else if (m_gamma == 2.f || m_gamma >= 12){
				// === Abremowitz and Stegun 15.3.5 + GSL maximum iteration error handling
				gsl_sf_result result;
				int status = gsl_sf_hyperg_2F1_e(b, c-a, c, normalZ, &result);
				double Fval = result.val;

				if(status){ // === GSL error handling
					if(status != GSL_EMAXITER)
						SLog(EError, "Microfacet model PGTR: uncaught GSL error (z=%g and normalZ=%g)",z,normalZ);
				}
				return pow(1.f-z,-b)*Fval;
			} else {
				// === Abremowitz and Stegun 15.3.4
				return pow(1.f-z,-a)*gsl_sf_hyperg_2F1(a,c-b,c,normalZ);
			}
		} else {
			return gsl_sf_hyperg_2F1(a,b,c,z);
		}
	}

protected:
	ETypeSTD m_type;
	Float m_alphaU, m_alphaV;
	bool m_sampleVisible;

	bool m_correlatedGAF;

	Float m_gamma;
	Float m_gammaMinOne;
	Float m_sqrtGammaMinOne;
	double m_F1;
	double m_F2;

	double m_S2_G1;
	double m_S2_G2;
};

MTS_NAMESPACE_END

#endif /* SRC_BSDFS_MICROFACETSTD_H_ */
